#!/usr/bin/env python
#
# Copyright (c) 2026 Authors and contributors
# (see the AUTHORS.rst file for the full list of names)
#
# Released under the GNU Public Licence, v3 or any higher version
# SPDX-License-Identifier: GPL-3.0-or-later
r"""Library modules of MAICoS.

This library contains additional modules, like general and mathematical helper
functions, which are used in the other MAICoS modules.
"""

__all__ = [
    "util",
    "math",
    "weights",
]
